"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const Contracts_1 = __importDefault(require("../../components/Contracts"));
const chai_1 = require("chai");
const ethers_1 = require("ethers");
const hardhat_1 = require("hardhat");
const merkletreejs_1 = __importDefault(require("merkletreejs"));
const { utils: { id, formatBytes32String, getAddress, solidityKeccak256, keccak256 }, constants: { AddressZero: ZERO_ADDRESS } } = hardhat_1.ethers;
const ROLE_GOVERNOR = id('ROLE_GOVERNOR');
const ROLE_MINTER = id('ROLE_MINTER');
describe('StakingRewardsClaim', () => {
    let deployer;
    let masterVault;
    let contractRegistry;
    let bancorNetworkV3;
    let bnt;
    let bntGovernance;
    before(async () => {
        [deployer] = await hardhat_1.ethers.getSigners();
    });
    beforeEach(async () => {
        bnt = await Contracts_1.default.DSToken.deploy('BNT', 'BNT', 18);
        bntGovernance = await Contracts_1.default.TestTokenGovernance.deploy(bnt.address);
        await bntGovernance.grantRole(ROLE_GOVERNOR, deployer.address);
        await bnt.transferOwnership(bntGovernance.address);
        await bntGovernance.acceptTokenOwnership();
        contractRegistry = await Contracts_1.default.ContractRegistry.deploy();
        bancorNetworkV3 = await Contracts_1.default.TestBancorNetworkV3.deploy(contractRegistry.address);
        bancorNetworkV3.setNetworkToken(bnt.address);
        masterVault = await Contracts_1.default.TokenHolder.deploy();
        bancorNetworkV3.setBancorVault(masterVault.address);
    });
    describe('construction', () => {
        const root = formatBytes32String('root');
        it('should revert when attempting to create with an invalid network contract', async () => {
            await (0, chai_1.expect)(Contracts_1.default.StakingRewardsClaim.deploy(ZERO_ADDRESS, bntGovernance.address, root)).to.be.revertedWith('InvalidAddress');
        });
        it('should revert when attempting to create with an invalid BNT governance contract', async () => {
            await (0, chai_1.expect)(Contracts_1.default.StakingRewardsClaim.deploy(bancorNetworkV3.address, ZERO_ADDRESS, root)).to.be.revertedWith('InvalidAddress');
        });
        it('should be properly initialized', async () => {
            const stakingRewardsClaim = await Contracts_1.default.StakingRewardsClaim.deploy(bancorNetworkV3.address, bntGovernance.address, root);
            (0, chai_1.expect)(await stakingRewardsClaim.merkleRoot()).to.equal(root);
        });
    });
    describe('rewards', () => {
        let stakingRewardsClaim;
        let provider1;
        let provider2;
        let provider3;
        let provider4;
        let provider5;
        let rewards = {};
        let merkleTree;
        const generateLeaf = (address, amount) => Buffer.from(solidityKeccak256(['address', 'uint256'], [getAddress(address), amount.toString()]).slice(2), 'hex');
        before(async () => {
            [provider1, provider2, provider3, provider4, provider5] = await hardhat_1.ethers.getSigners();
            rewards = {
                [provider1.address]: ethers_1.BigNumber.from(1),
                [provider2.address]: ethers_1.BigNumber.from(100000),
                [provider3.address]: ethers_1.BigNumber.from(500),
                [provider4.address]: ethers_1.BigNumber.from(1000000),
                [provider5.address]: ethers_1.BigNumber.from(5)
            };
            merkleTree = new merkletreejs_1.default(Object.entries(rewards).map(([provider, amount]) => generateLeaf(provider, amount)), keccak256, { sortPairs: true });
        });
        beforeEach(async () => {
            stakingRewardsClaim = await Contracts_1.default.StakingRewardsClaim.deploy(bancorNetworkV3.address, bntGovernance.address, merkleTree.getRoot());
            await bntGovernance.grantRole(ROLE_MINTER, stakingRewardsClaim.address);
        });
        const testClaimOrStake = (stake) => {
            const claim = async (provider, overrides = {}) => {
                let { caller = provider, amount, proof } = overrides;
                if (!amount) {
                    amount = rewards[provider.address];
                }
                if (!proof) {
                    proof = merkleTree.getHexProof(generateLeaf(provider.address, amount));
                }
                return stake
                    ? stakingRewardsClaim.connect(caller).stakeRewards(provider.address, amount, proof)
                    : stakingRewardsClaim.connect(caller).claimRewards(provider.address, amount, proof);
            };
            describe(stake ? 'staking' : 'claiming', () => {
                it('should revert when calling for a different provider', async () => {
                    await (0, chai_1.expect)(claim(provider1, { caller: provider2 })).to.be.revertedWith('AccessDenied');
                });
                it('should revert when calling with a different amount', async () => {
                    await (0, chai_1.expect)(claim(provider1, { amount: rewards[provider2.address] })).to.be.revertedWith('InvalidClaim');
                });
                it('should revert when calling with an invalid proof', async () => {
                    const proof = merkleTree.getHexProof(generateLeaf(provider1.address, rewards[provider1.address]));
                    // incorrect provider
                    await (0, chai_1.expect)(claim(provider2, { proof })).to.be.revertedWith('InvalidClaim');
                    // incorrect proof
                    await (0, chai_1.expect)(claim(provider1, {
                        proof: [formatBytes32String('1234'), ...proof]
                    })).to.be.revertedWith('InvalidClaim');
                    // incorrect proof
                    await (0, chai_1.expect)(claim(provider1, {
                        proof: merkleTree.getHexProof(generateLeaf(provider2.address, rewards[provider2.address]))
                    })).to.be.revertedWith('InvalidClaim');
                    // incorrect proof
                    await (0, chai_1.expect)(claim(provider1, {
                        amount: rewards[provider2.address],
                        proof: merkleTree.getHexProof(generateLeaf(provider2.address, rewards[provider2.address]))
                    })).to.be.revertedWith('InvalidClaim');
                    // incorrect proof - partial path
                    await (0, chai_1.expect)(claim(provider1, { proof: [...proof].slice(1) })).to.be.revertedWith('InvalidClaim');
                });
                it('should revert when attempting to claim twice', async () => {
                    await claim(provider1);
                    await (0, chai_1.expect)(claim(provider1)).to.be.revertedWith('AlreadyClaimed');
                });
                it('should claim the rewards', async () => {
                    let totalClaimed = ethers_1.BigNumber.from(0);
                    for (const [providerAddress, amount] of Object.entries(rewards)) {
                        const provider = await hardhat_1.ethers.getSigner(providerAddress);
                        (0, chai_1.expect)(await stakingRewardsClaim.totalClaimed()).to.equal(totalClaimed);
                        (0, chai_1.expect)(await stakingRewardsClaim.hasClaimed(providerAddress)).to.be.false;
                        const prevTotalSupply = await bnt.totalSupply();
                        const prevProviderBalance = await bnt.balanceOf(providerAddress);
                        const prevVaultBalance = await bnt.balanceOf(masterVault.address);
                        const res = await claim(provider);
                        await (0, chai_1.expect)(res)
                            .to.emit(stakingRewardsClaim, stake ? 'RewardsStaked' : 'RewardsClaimed')
                            .withArgs(providerAddress, amount);
                        totalClaimed = totalClaimed.add(amount);
                        (0, chai_1.expect)(await stakingRewardsClaim.totalClaimed()).to.equal(totalClaimed);
                        (0, chai_1.expect)(await stakingRewardsClaim.hasClaimed(providerAddress)).to.be.true;
                        (0, chai_1.expect)(await bnt.totalSupply()).to.equal(prevTotalSupply.add(amount));
                        (0, chai_1.expect)(await bnt.balanceOf(providerAddress)).to.equal(prevProviderBalance.add(stake ? 0 : amount));
                        // please note the "real" master vault balance shouldn't change when users stake, but this is
                        // how the system should behave during these tests since we're stubbing the master vault with q
                        // TokenHolder contract
                        (0, chai_1.expect)(await bnt.balanceOf(masterVault.address)).to.equal(prevVaultBalance.add(stake ? amount : 0));
                    }
                });
            });
        };
        for (const stake of [true, false]) {
            testClaimOrStake(stake);
        }
    });
});
//# sourceMappingURL=StakingRewardsClaim.js.map